table Imu {
    t: int64;	
    temperature: float;
    accelerometer_x: float;
    accelerometer_y: float;
    accelerometer_z: float;
    gyroscope_x: float;
    gyroscope_y: float;
    gyroscope_z: float;
    magnetometer_x: float;
    magnetometer_y: float;
    magnetometer_z: float;
}

table ImuPacket {
    elements: [Imu];
}

root_type ImuPacket;